/*   video test program for the AM-RGB module  
 *   free public release
 *
 *   Rin. Taniguchi       email: rin@limu.is.kyushu-u.ac.jp
 *
 *   Dept. of Intelligent Systems
 *   Kyushu University
 *   Japan
 *
 *   test_icpci_rgb.c  v1.0  1998/01/05
 */

/* Change History:

   1999/02/19 14:13:37  matthias
   changes to match the new on-board/upper memory mapping scheme (untested)
   struct icp_info is only available in bma-mode

   1999/03/02 10:24:51  matthias
   minor adjustments to avoid egcs111 warnings
   new glibc has 1024 signals / libc had only 32 (ifdef __GLIBC__)

   1999/05/03 13:20:45  matthias
   ajustments to version 0.5.0 (icpciIoctl.h > icpciExt.h)
   card 8/16/32 -> u_char/short/long
   
   1999/09/21 15:51:16  matthias
   ajustments for version 0.5.1
   the statistics are now u_long values
   due to the capability of decimazation we have to distinguish between
   MEMWIDTH/HEIGHT and BMAWIDTH/HEIGHT
   RAWSIZE is now MEMBUFSIZE ans PAGEDSIZE is now BMABUFSIZE
   MEMSIZE is renamed to MEMMAPSIZE

 */


/*
 * Defines:
 */

/*
 * Includes:
 */

#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include <string.h>
#include <signal.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/mman.h>

#include <asm/page.h>

#include "icpciExt.h"

#define  EXOFILE "amstd.exo"
#define  FRAMES 20

typedef enum {
  READ_MODE,
  MMAP_MODE,
  BMA_MODE
} operation_mode_t;

typedef enum {
  GRAB_MODE,
  SNAP_MODE,
  FREEZE_MODE
} acquisition_mode_t;

/*
 * Devices:
 */

const char *rgb_camera_device[] = {
  "/dev/ic0rgb0" , "/dev/ic0rgb1" , 
  "/dev/ic1rgb0" , "/dev/ic1rgb1" ,
  "/dev/ic2rgb0" , "/dev/ic2rgb1" ,
  "/dev/ic3rgb0" , "/dev/ic3rgb1"
};

const char *rgb_ioctl_device[] = {
  "/dev/ic0rgbI",
  "/dev/ic1rgbI",
  "/dev/ic2rgbI",
  "/dev/ic3rgbI"
};

const char *rgb_ident_device[] = {
  "/dev/ic0idI",
  "/dev/ic1idI",
  "/dev/ic2idI",
  "/dev/ic3idI" 
};

/*
 * Globals:
 */

void	       *image_mmaped = NULL;	/* pointer to mmapped frame buffer
					   (on-board or upper memory) */

size_t		mem_size = 0UL;		/* size of on-board/uppen memory */


int framegrabber=0;
operation_mode_t        operation_mode = BMA_MODE;
acquisition_mode_t      acquisition_mode = GRAB_MODE;


/*
 * Prototypes:
 */

/*
 * int ascii2decimal (char *)
 *
 * convert tow  characters, which are following another and form a
 * hexadecimal number into a decimal value
 */

static int 
ascii2decimal (char *pos) {

  int value;

  /* convert ASCII character to the first digit of a decimal value */

  if ((*pos >= '0') && (*pos <= '9'))
    value = *pos++ - '0';
  else if ((*pos >= 'A') && (*pos <= 'F'))
    value = *pos++ - 'A' + 10;
  else if ((*pos >= 'a') && (*pos <= 'f'))
    value = *pos++ - 'a' + 10;
  else
    return (-1);

  /* convert ASCII character to the second digit of a decimal value */

  if ((*pos >= '0') && (*pos <= '9'))
    value = (value << 4) + *pos - '0';
  else if ((*pos >= 'A') && (*pos <= 'F'))
    value = (value << 4) + *pos - 'A' + 10;
  else if ((*pos >= 'a') && (*pos <= 'f'))
    value = (value << 4) + *pos - 'a' + 10;
  else
    return (-1);

  return value;
}

/*
 * int rgb_fpga_load (char *)
 *
 * read the FPGA-Timing from the proper ASCII-File with error checking,
 * calculate for every line the checksum and convert it into binary format
 */

static int rgb_fpga_load (char *filename, struct fpga_buffer_t *fpga_buffer) {

  FILE *file;
  int number, high, mid, low, data;
  int line_count, data_count, fpga_count=0;
  char *home, *pos, line[256];
  unsigned char checksum;

  /* get path of mcs file for open */

  if ((home = getenv ("ICPCILIB")) == NULL) {
    fprintf (stderr, "can't find environment variable 'ICPCILIB'.\n");
    return (-1);
  }

  strcpy (line, home);
  strcat (line, "/");
  strcat (line, filename);

  if ((file = fopen (line, "r")) == NULL) {
    fprintf (stderr, "can't open '%s'.\n", line);
    return (-1);
  }

  /* read mcs file and do error checking */

  for (line_count=0; fgets (line, 256, file) != NULL; line_count++) { 
    checksum = 0;
    pos = line;

    if (*pos == '#') continue;
    if (strncmp(pos, "S9", 2) == 0) break;
    if (strncmp(pos, "S2", 2) != 0) {
      fprintf (stderr, "can't find S2 nor S9 in file '%s' at line %d.\n", 
        filename, line_count);
      return (-1);
    } 
    else pos += 2;

    /* read number of data bytes */

    if ((number = ascii2decimal (pos)) < 0) {    
      fprintf (stderr, "illegal character found in file '%s' at line %d.\n",
        filename, line_count);
      return (-1);
    }
    else {   
      pos += 2;     
      checksum += (unsigned char) number;
    }

    /* read high byte of addres */

    if ((high = ascii2decimal (pos)) < 0) {    
      fprintf (stderr, "illegal character found in file '%s' at line %d.\n",
        filename, line_count);
      return (-1);
    }
    else {
      pos += 2;
      checksum += (unsigned char) high;
      number--;
    }

    /* read mid byte of addres */

    if ((mid = ascii2decimal (pos)) < 0) {    
      fprintf (stderr, "illegal character found in file '%s' at line %d.\n",
        filename, line_count);
      return (-1);
    }
    else {
      pos += 2;
      checksum += (unsigned char) mid;
      number--;
    }

    /* read low byte of addres */

    if ((low = ascii2decimal (pos)) < 0) {    
      fprintf (stderr, "illegal character found in file '%s' at line %d.\n",
        filename, line_count);
      return (-1);
    }
    else {
      pos += 2;
      checksum += (unsigned char) low;
      number--;
    }

    number--; /* pre-decrement for check sum */

    /* read data bytes */
   
    for (data_count=0; data_count<number; data_count++) {
      if ((data = ascii2decimal (pos)) < 0) {    
        fprintf (stderr, "illegal character found in file '%s' at line %d.\n",
          filename, line_count);
        return (-1);
      }
      else {
	if (fpga_count > RGB_FPGA_MAX_SIZE) {
	  fprintf (stderr, "Too large fpga data\n");
	  return (-1);
	}
        fpga_buffer->buffer[fpga_count++] = data;
        pos += 2;
        checksum += (unsigned char) data;
      }
    }

    /* read checksum byte */
 
    if ((data = ascii2decimal (pos)) < 0) {    
      fprintf (stderr, "illegal character found in file '%s' at line %d.\n",
        filename, line_count);
      return (-1);
    }
    else if ((unsigned char)(~checksum) != (unsigned char)data) {
      fprintf (stderr, "checksum error in file '%s' at line %d.\n",
        filename, line_count);   
      return (-1);
    }
  }

  fclose (file);
  fpga_buffer->size = fpga_count;
  return (0);
}


void
Clean_Up (void) {

  if (image_mmaped != NULL && mem_size != 0) {
    munmap (image_mmaped, mem_size);
  }

  if (framegrabber != -1) {
    close (framegrabber);
  }
  
  exit (-1);
}

void write_image_to_file
 (int frame_count, u_long *read_buffer, int width, int height) {

  char str[64];
  int j, k, o;
  u_char *buf;

  sprintf(str, "im.%03d.ppm", frame_count);
  fprintf(stderr, "Writing image %d\n", frame_count);
  if ((o = open(str, O_WRONLY | O_CREAT, 0644)) < 0) {
    perror("ppm open failed");
    Clean_Up();
  }
  
  /* make PPM header and save to file */
  sprintf(str, "P6\n%d %d\n255\n", 640, 480);
  write (o, &str[0], strlen(str));
  /* save the data to PPM file */

  buf = (u_char *)read_buffer;
  for (j=0; j<480; j++) {
    for (k=0; k<640; k++) {
      write(o, buf, 3);
      buf+=4;
    }
  }
  
  close(o);
}

void write_interlace_image_to_file
 (int frame_count, u_long *read_buffer, int width, int height) {

  char str[64];
  int j, k, o;
  u_char *buf0, *buf1;

  sprintf(str, "im.%03d.ppm", frame_count);
  fprintf(stderr, "Writing image %d\n", frame_count);
  if ((o = open(str, O_WRONLY | O_CREAT, 0644)) < 0) {
    perror("ppm open failed");
    Clean_Up();
  }
  
  /* make PPM header and save to file */
  sprintf(str, "P6\n%d %d\n255\n", 640, 480);
  write (o, &str[0], strlen(str));
  /* save the data to PPM file */

  buf0 = (u_char *)read_buffer;
  buf1 = (u_char *)read_buffer + 640 * 240 *4;
  for (j=0; j<240; j++) {
    for (k=0; k<640; k++) {
      write(o, buf0, 3);
      buf0+=4;
    }
    for (k=0; k<640; k++) {
      write(o, buf1, 3);
      buf1+=4;
    }
  }
  
  close(o);
}


/*
 * Main:
 */

int main(int argc, char **argv) {


  fd_set wait_fdset;
  struct timeval wait_time;
  int frame_count;

  struct icp_info_t *icp_info = NULL;
  struct fpga_buffer_t fpga_buffer;

  int board_arg = 0, camera_arg = 0;

  int width=0, height=0, rawsize=0, pagedsize=0;
  u_long *read_buffer=NULL;

  static struct sigaction ignore_action = { SIG_IGN,
#ifdef __GLIBC__
					    { { 0, 0, 0, 0 } },
#else
					    0,
#endif
					    0,
					    NULL };

  switch (operation_mode) {
  case READ_MODE:
    if ((framegrabber = open (rgb_camera_device[board_arg * 4 + camera_arg],
                              O_RDONLY)) < 0) {
      perror ("open framegrabber device (with AM-CLR)");
      exit (-1);
    }
    break;

  case MMAP_MODE:
    if ((framegrabber = open (rgb_ioctl_device[board_arg],
			    O_RDONLY)) < 0) {
      perror ("open camera device");
      exit (-1);
    }
    break;

  case BMA_MODE:
    if ((framegrabber = open (rgb_ident_device[board_arg],
			    O_RDONLY)) < 0) {
      perror ("open camera device");
      exit (-1);
    }
  }

  /* get the current loaded FPGA-Timing */


  /* check if an FPGA-Timing was given at command line */
  /* read the new MCS-Timing File */
  if (rgb_fpga_load (EXOFILE, &fpga_buffer) < 0) {
    fprintf (stderr, "can't load mxo file.\n");
    Clean_Up();
  }
  printf(" size = %d\n", fpga_buffer.size);

  /* write the new Timing into FPGA-Chip */
  if (ioctl (framegrabber, HIOC_RGB_SET_FPGA, &fpga_buffer) < 0) {
    perror ("GIOC_RGB_SET_FPGA");
    Clean_Up();
  }
  fprintf(stderr, "Set FPGA OK\n");


  if (operation_mode != READ_MODE) {
    /* MMAP_MODE or BMA_MODE */
    if (sigaction (SIGIO, &ignore_action, NULL) < 0) {
      perror ("Can't set handler to ignore action");
      Clean_Up ();
    }
    if (fcntl (framegrabber, F_SETOWN, getpid()) < 0) {
      perror ("F_SETOWN");
      Clean_Up ();
    }
    if (fcntl (framegrabber, F_SETFL,
	       fcntl (framegrabber, F_GETFL, NULL) | FASYNC) < 0) {
      perror ("F_SETFL FASYNC");
      Clean_Up ();
    }
  }

  /* --- get image dimensions --------------------------------------------- */

  switch (operation_mode) {
  case READ_MODE:
  case MMAP_MODE:
    if (ioctl (framegrabber, GIOC_GE_GET_MEMWIDTH, &width) < 0) {
      perror ("GIOC_GE_GET_WIDTH");
      Clean_Up();
    }
    if (ioctl (framegrabber, GIOC_GE_GET_MEMHEIGHT, &height) < 0) {
      perror ("GIOC_GE_GET_HEIGHT");
      Clean_Up();
    }
    break;
    
  case BMA_MODE:
    if (ioctl (framegrabber, GIOC_GE_GET_BMAWIDTH, &width) < 0) {
      perror ("GIOC_GE_GET_WIDTH");
      Clean_Up();
    }
    if (ioctl (framegrabber, GIOC_GE_GET_BMAHEIGHT, &height) < 0) {
      perror ("GIOC_GE_GET_HEIGHT");
      Clean_Up();
    }
    break;
  }

  if (ioctl (framegrabber, GIOC_GE_GET_MEMBUFSIZE, &rawsize) < 0) {
    perror ("GIOC_GE_GET_RAWSIZE");
    Clean_Up();
  }
  if (ioctl (framegrabber, GIOC_GE_GET_BMABUFSIZE, &pagedsize) < 0) {
    perror ("GIOC_GE_GET_PAGEDSIZE");
    Clean_Up();
  }

  /* --- initialize the given operation mode ------------------------------ */

  switch (operation_mode) {
  case READ_MODE:
    if ((read_buffer = (u_long *) malloc (rawsize)) == NULL) {
      perror ("can't malloc read buffer.\n");
      Clean_Up();
    }
    break;
  case MMAP_MODE:
    if (ioctl (framegrabber, GIOC_GE_GET_MEMMAPSIZE, &mem_size) < 0) {
      perror ("GIOC_GE_GET_MEMSIZE");
      exit (-1);
    }
    if ((image_mmaped = (u_long*)mmap (0, mem_size, PROT_READ, MAP_SHARED,
				       framegrabber , 0)) == (void*)-1) {
      perror ("Can't map camera device");
      exit (-1);
    }
    break;
    
  case BMA_MODE:
    if (ioctl (framegrabber, GIOC_GE_GET_BMAMAPSIZE, &mem_size) < 0) {
      perror ("GIOC_GE_GET_BMASIZE");
      exit (-1);
    }
    if ((image_mmaped = (void*)mmap (0, mem_size, PROT_READ, MAP_SHARED,
				     framegrabber, 0)) == (void*)-1) {
      perror ("Can't map camera device");
      exit (-1);
    }
    /* pointer to the info structure of the acquired image */
    icp_info = (struct icp_info_t *)(((u_char *)image_mmaped) + rawsize);
    break;
  }

  if (operation_mode != READ_MODE) {
    /* MMAP_MODE or BMA_MODE */
    if (ioctl (framegrabber, GIOC_CR_SET_GRAB, NULL) < 0) {
      perror ("GIOC_CR_SET_GRAB");
      Clean_Up();
    }
  }

  frame_count = 0;
  while (frame_count < FRAMES) {

    switch (acquisition_mode) {

    case GRAB_MODE:
      switch(operation_mode) {
   
      case READ_MODE:
        if (read (framegrabber, read_buffer, rawsize) < 0) {
          if (errno == ENODATA) {
	    fprintf(stderr, "*** No Sync ***");
	  }
          else {
            perror ("read framegrabber device");
	    Clean_Up();
          }        
        }

        write_image_to_file (frame_count++, read_buffer, width, height);
        break;

      case MMAP_MODE:
        FD_ZERO (&wait_fdset);
        FD_SET (framegrabber, &wait_fdset);	    
        wait_time.tv_sec = 1;
        wait_time.tv_usec = 0;
        switch (select (getdtablesize(), &wait_fdset, NULL, NULL, 
                        &wait_time)) {
        case -1:
	  fprintf(stderr, "*** Signal in select ***\n");
          break;
        case 0:
          fprintf(stderr, "*** No Sync ***\n");
          break;
        case 1:
	  write_interlace_image_to_file (frame_count++, image_mmaped,
					 width, height);
	  break;
        default:
          perror ("select got unexpected return value");
	  Clean_Up();
        }
        break;

      case BMA_MODE:
        FD_ZERO (&wait_fdset);
        FD_SET (framegrabber, &wait_fdset);	    
        wait_time.tv_sec = 1;
        wait_time.tv_usec = 0;
        switch (select (getdtablesize(), &wait_fdset, NULL, NULL, 
                        &wait_time)) {
        case -1:
	  fprintf(stderr,"*** Signal in select ***\n");
          break;
        case 0:
          fprintf(stderr, "*** No Sync ***\n");
          break;
        case 1:
	  write_image_to_file (frame_count++, image_mmaped, width, height);
	  break;
        default:
	  perror ("select got unexpected return value");
	  Clean_Up();
        }
        break;
  
      default:
        perror("unknown acquisition mode.\n");
	Clean_Up();
      }
      break;
       
 
    default:
      perror("unknown operation mode.\n");
      Clean_Up();
    }

  }

  if (operation_mode != READ_MODE) {
    /* MMAP_MODE or BMA_MODE */
    if (ioctl (framegrabber, GIOC_CR_SET_FREEZE, NULL) < 0) {
      perror ("GIOC_CR_SET_FREEZE");
      Clean_Up();
    }
    if (operation_mode == BMA_MODE) {
      /* output info of last acquired image */
      printf ("clock tics: %ld\n", icp_info->clock_tics);
      printf ("time stamp: %d sec, %d usec\n", 
	      (int)icp_info->time_stamp.tv_sec,
	      (int)icp_info->time_stamp.tv_usec);
      printf ("frames captured: %ld\n", icp_info->frame_nums.frames_captured);
      printf ("   frames valid: %ld\n", icp_info->frame_nums.frames_valid);
      printf ("    frames lost: %ld\n", icp_info->frame_nums.frames_lost);
    }
  }

  if (read_buffer != NULL) free (read_buffer);

  if (operation_mode != READ_MODE) {
    if (munmap (image_mmaped, mem_size) < 0) {
      perror ("unmap framebuffer");
      exit (-1);
    }
  } 

  if (close (framegrabber) < 0) {
    perror ("close framegrabber device");
    exit (-1);
  }

  return (0);
}
 

